#!/usr/bin/perl -w

# displog.cgi - format and display your web server's log file
# V1.5 August 3, 2003
# If you want details only available in your log file, this program can help
# Features:
#   Displays log file path and server's time and date
#   Skips over unwanted log entries - displays from any date to the end of the file
#   Optionally formats IP address, date/time, filename, referrer in columns
#   Groups sessions by color indicating the 1st file visitors viewed, how long they stayed and how many files they viewed
#   Reduces clutter by optionally skipping *.jpg, *.gif, *.css files, attempted virus attacks, and cgi parameters
#   Hilights referrals from search engines
#		Three optional keys match and display only selected log entries
#   WILL NOT CHANGE YOUR LOG FILE OR ANYTHING ELSE - It just reads the log and displays the requested information

# This program requires that your log file use the format:
# 63.93.81.10 - - [17/Apr/2002:09:44:47 -0600] "GET /index.shtml HTTP/1.0" 200 24173 "[unknown origin]" "Mozilla/4.78 [en] (Win98; U)"
#	IP address<" "> [dd/mmm/yyyy:hh:mm:ss      ] "Command/filename"                    "Referred From"    "Browser"
#		IP address starting in column 0 and followed by a space
#   date/time 20 characters long enclosed by []
#   command/filename enclosed by ""
#   referring website enclosed by ""
#   browser enclosed by ""

# This unsupported, public domain perl script was written by:
# Don Rowe
# http://www.canzonatech.com
# This program is provided as-is at no charge and is guaranteed to be worth at least half that.
# Just set the perl path (above), log file name ($filename below), upload with FTP and chmod to 755.
# Optional command line parameters are:
#		'days' overrides $BackupDays
#		'fmt' overrides $FormatText
#		'all' overrides $ShowJpgFiles, $ShowGifFiles, $ShowCssFiles, $ShowVirusFiles, $ShowCgiParams
#		'key1'-'key3' are case-insensitive search fields.  If specified, only matching log entries display
#		'and' overrides $LogicAnd
#		'refer' overrides $ReferOnly
#		'fmt=0&all=1' displays everything for the requested number of days
# Delays of several seconds are normal while the program scans your log file.
# If the file runs but you don't see anything at all, check the file name and path as it didn't open
# Please send suggestions, fixes, and enhancements so others may share

# Version history
# V1.2
#		always displays parameters for filetext.cgi
# V1.3
#		search/query also identified by "result"
# V1.4
#		Referer string displays as link if not $domainname
# V1.5
#		$domainname must be at beginning of Referer string
#		refer parameter added to display only external referrers

#-----------------------
# Required Configuration
#-----------------------

# These are the only options you MUST change

# Full path to your log file
# Check with your website host if you're not sure
$filename = "/home/log/weblog";

# Domain name
$domainname = "yourdomain.com";

#-----------------------
# Optional Configuration
#-----------------------

# Number of days (from end) to show if not specified
$BackupDays = 7;

# 1=Show *.jpg file accesses, 0=don't
$ShowJpgFiles = 0;

# 1=Show *.gif file accesses, 0=don't
$ShowGifFiles = 0;

# 1=Show *.css file accesses, 0=don't
$ShowCssFiles = 1;

# 1=Show attempted virus attacks, 0=don't
$ShowVirusFiles = 0;

# 1=Show cgi options after '?', 0=don't
$ShowCgiParams = 0;

# Server's offset in hours to GMT (GMT-local time)
$GmtOffset = 7;

# Default font (column formatting works best with fixed-width fonts)
$FontFace = "Fixedsys";

# Text color for IP address grouping
$FontColor = "green";

# Search/query string color
$SearchColor = "red";

# Day grouping background color
$DayColor = "#FFFF99";

# Padded width of IP address field
$PadIp = 16;

# Padded width of Date/Time field
$PadDate = 22;

# Padded width of filename
$PadFile = 40;

# 1=Reformat line into columns, 0=display entire line as-is
$FormatText = 1;

# 1=AND all search keys, 0=OR all search keys
$LogicAnd = 0;

# 1=Show external referrers only, 0=all (normal)
$ReferOnly = 0;

#---------------------
# End of configuration
#---------------------

@aKeys = ();					# No match keys specified
@DaysInMonth = (31,28,31,30,31,30,31,31,30,31,30,31);
$MonthNames = "janfebmaraprmayjunjulaugsepoctnovdec";

if ($ENV{'REQUEST_METHOD'} eq "GET")
{
	$CmdParams = $ENV{'QUERY_STRING'};
}
else
{
	read(STDIN, $CmdParams, $ENV{'CONTENT_LENGTH'});
}                                                         
@aCmdParams = split (/&/,$CmdParams);
%hCmdParams = ();				# Empty for now
foreach $pair (@aCmdParams) {                 
    ($key, $value) = split (/=/,$pair);
    $value =~ tr/+/ /;
    $value =~ s/%([\dA-Fa-f][\dA-Fa-f])/pack ("C", hex ($1))/eg;
    $hCmdParams{$key} = $value;  # store the key in the results hash
}

if (exists $hCmdParams{"days"})
{
	# Set number of days to display
	$BackupDays = $hCmdParams{"days"};
}

if (exists $hCmdParams{"fmt"})
{
	# formatting
	$FormatText = $hCmdParams{"fmt"};
}

if (exists $hCmdParams{"and"})
{
	# formatting
	$LogicAnd = $hCmdParams{"and"};
}

if (exists $hCmdParams{"refer"})
{
	# formatting
	$ReferOnly = $hCmdParams{"refer"};
}

if (exists $hCmdParams{"key1"})
{
	# formatting
	if (length($hCmdParams{"key1"}) > 0)
	{
		$aKeys[0] = $hCmdParams{"key1"};
	}
}

if (exists $hCmdParams{"key2"})
{
	# formatting
	if (length($hCmdParams{"key2"}) > 0)
	{
		$aKeys[1] = $hCmdParams{"key2"};
	}
}

if (exists $hCmdParams{"key3"})
{
	# formatting
	if (length($hCmdParams{"key3"}) > 0)
	{
		$aKeys[2] = $hCmdParams{"key3"};
	}
}

if (exists $hCmdParams{"all"})
{
	# Show all optional files
	$ShowJpgFiles = $hCmdParams{"all"};
	$ShowGifFiles = $hCmdParams{"all"};
	$ShowCssFiles = $hCmdParams{"all"};
	$ShowVirusFiles = $hCmdParams{"all"};
	$ShowCgiParams = $hCmdParams{"all"};
}

# Set starting date
$StartDate = time-($BackupDays * (24*60*60));

print "Content-type: text/html\n\n";

if (open (FILE, "$filename"))
{
	print qq~
	<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2//EN">
	<HTML><BODY>
	<TABLE BORDER="0" CELLPADDING="0" CELLSPACING="0" WIDTH="100%"><TR><TD WIDTH="100%">
	<FONT FACE="$FontFace">
	~;
	$text = localtime(time);
	print "$text - Last $BackupDays days of $filename.<BR>\r\n";

	$DateMatch = 0;												# Match flag for formatting
	$OldIp = "0.0.0.0";										# Old IP address for color grouping
	$IpColor = 0;													# Color grouping flag determines current color
	$OldDate = "00";											# Old date for color grouping
	$DateColor = 0;												# Color grouping flag determines day color
	$KeyMatch = 1;												# Assume match
	$DomainPos = 0;												# Position of domain name
	$SlashPos = 0;												# Position of 1st slash in referrer
	while (!eof(FILE))
	{
	  $text = <FILE>;
		@ip = split (/ /, $text, 2);				#$ip[0] holds IP address

		# Find start of date/time
		@date = split(/\[/, $text, 2);			#$date[1] starts with date/time
		$TextDate = substr($date[1], 0, 20);
		$TextDay = substr($TextDate, 0, 2);
		$TextMon = substr($TextDate, 3, 3);
		$TextYear = substr($TextDate, 7, 4);
		$TextTime = substr($TextDate, 12, 8);

		# Once a match, always a match
		if ($DateMatch == 0)
		{
			$secs = &ElapsedSecs(&MonthIndex($TextMon), $TextDay, $TextYear);
			if ($secs >= $StartDate)
			{
				$DateMatch = 1;
			}
		}

		if ($DateMatch > 0)
		{
			# Find command/filename
			$TextCmd = "-";
			$TextFile = "-";
			$TextType = "-";
			@quotes = split(/\"/, $text);
			if ($#quotes > 0)
			{
				@cmd = split(/ /, $quotes[1]);
				$TextCmd = $cmd[0];
				if ($#cmd > 0)
				{
					@file = split(/\./, $cmd[1]);
					$TextFile = $cmd[1];
					if (!$ShowCgiParams)
					{
						# Supress cgi parameters after the '?'
						$question = index($TextFile, "?");
						if ($question > 0)
						{
							if (index($TextFile, "filetext.cgi") < 0)
							{
								# Always show parameters for filetext.cgi
								$TextFile = substr($TextFile, 0, $question);
							}
						}
					}
					if ($#file > 0)
					{
						$TextType = lc($file[1]);
					}
				}
			}
			# Set referrer and browser
			$TextRefer = "-";		# May not be present
			$TextBrowser = "-";	# May not be present
			if ($#quotes >= 5)
			{
				$TextRefer = $quotes[3];
				$TextBrowser = $quotes[5];
			}

			if (!$ShowVirusFiles)
			{
				if
				(
					(index($TextFile, "default.ida") >= 0) ||
					(index($TextFile, "root.exe") >= 0) ||
					(index($TextFile, "cmd.exe") >= 0)
				)
				{
					# Suspend display until IP address changes
					$DateMatch = 2;
					$OldIp = $ip[0];
				}
				else
				{
					if ($DateMatch == 2)
					{
						if ($OldIp ne $ip[0])
						{
							# IP changed, resume normal display
							$DateMatch = 1;
						}
					}
				}
			}
			if ($DateMatch == 1)
			{
				$KeyMatch = 1;
				if ($#aKeys >= 0)											# Always a match if no keys defined
				{
					$KeyMatch = 0;											# Assume no match
					foreach $match (@aKeys)
					{
						if ($text =~ m/$match/i)
						{
							$KeyMatch++;
							if (!$LogicAnd) {last;}						# No need to keep searching
						}
					}
					if ($LogicAnd)
					{
						if ($KeyMatch < ($#aKeys+1))
						{
							$KeyMatch = 0;									# All or nothing
						}
					}
				}
			}

			if (($DateMatch == 1) && $KeyMatch)
			{
				if ($ShowGifFiles || ($TextType ne "gif"))
				{
					if ($ShowJpgFiles || ($TextType ne "jpg"))
					{
						if ($ShowCssFiles || ($TextType ne "css"))
						{
							if ($OldDate ne $TextDay)
							{
								$DateColor++;
								print "</FONT></TD></TR><TR>";
								if ($DateColor & 1)
								{
									print qq~<TD WIDTH="100%" BGCOLOR="$DayColor">~;
								}
								else
								{
									print qq~<TD WIDTH="100%">~;
								}
								&SetFontColor($IpColor & 1);
								$OldDate = $TextDay;
							}
							if ($OldIp ne $ip[0])
							{
								$IpColor++;
								&SetTextColor($IpColor & 1);
								$OldIp = $ip[0];
							}
							if ($TextRefer eq "-")
							{
								$DomainPos = 0;
							}
							else
							{
								$DomainPos = index($TextRefer, $domainname);
								$SlashPos = index($TextRefer, "/", 7);
								if ($SlashPos > 0)
								{
									if ($DomainPos > $SlashPos)
									{
										$DomainPos = -1;
									}
								}
							}
							if (!$ReferOnly || ($DomainPos < 0))
							{
								if ($FormatText)
								{
									# Format in columns
									&PadString($ip[0], $PadIp);
									&PadString($TextDate, $PadDate);
									&PadString($TextFile, $PadFile);

									if ((index(lc($TextRefer), "query") >= 0) || (index(lc($TextRefer), "search") >= 0) || (index(lc($TextRefer), "result") >= 0) || (index(lc($TextRefer), "answer") >= 0))
									{
										print qq~</FONT><FONT FACE="$FontFace" COLOR="$SearchColor">~;
										print "$TextRefer";
										&SetTextColor($IpColor & 1);
									}
									else
									{
										if ($DomainPos < 0)
										{
											print qq~<A HREF="~;
											print "$TextRefer";
											print qq~">$TextRefer</A>~;
										}
										else
										{
											print "$TextRefer";
										}
									}
								}
								else
								{
									# Display entire line as-is
									print "$text";
								}
								print "<BR>\r\n";
							}
						}
					}
				}
			}
		}
	}
	print "</FONT></TD></TR></TABLE></BODY></HTML>\r\n";
	close (FILE);
	exit;
}

# Pad and display text(string, pad width)
sub PadString
{
	$tx = $_[0];		# string to output
	$len = $_[1];		# length
	$ltx = length($tx);
	print "$tx";
	for ($i = $ltx; $i < $len; $i++)
	{
		print "&nbsp;"
	}
}

# Set font color (0 selects normal, !0 selects alternate)
sub SetFontColor
{
	if ($_[0])
	{
		# Alternate font/color
		print qq~<FONT FACE="$FontFace" COLOR="$FontColor">~;
	}
	else
	{
		# Normal font/color
		print qq~<FONT FACE="$FontFace">~;
	}
}

# Cancel current font and set new color (0 selects normal, !0 selects alternate)
sub SetTextColor
{
	$NewColor = $_[0];
	print "</FONT>";
	SetFontColor($NewColor);
}

# Convert month string to [1-12] index(3-character month abbreviation)
# Invalid month name returns 0
sub MonthIndex
{
	$month = lc(substr($_[0], 0, 3));
	$result = index($MonthNames, $month);
	if ($result < 0)
	{
		# Not found
		$result = 0;
	}
	else
	{
		if ($result % 3)
		{
			# strange (but invalid) character match
			$result = 0;
		}
		else
		{
			# This is a keeper
			$result = ($result / 3) + 1;
		}
	}
	return $result;
}

# Convert date values to seconds since 1Jan1970 GMT(month[1-12], date[1-31], year[00-99 or 1970-])
sub ElapsedSecs
{
	$month = $_[0];
	$date = $_[1];
	$year = $_[2];

	$result = $GmtOffset*60*60;

	if ($year < 100)
	{
		# 2 digit year
		$year = $year + 1900;
		if ($year < 1970)
		{
			# Assume 2000-2069
			$year += 100;
		}
	}

	# Seconds in previous years
	$result += 365*24*60*60 * ($year - 1970);
	# Adjust for intervening leap years (1st was 1972)
	$result += (24*60*60 * (($year-1969) >> 2));

	# Seconds in previous months
	for ($i = 1; $i < $month; $i++)
	{
		$result += (24*60*60 * $DaysInMonth[$i-1]);
	}
	# Adjust for leap year
	if (($month > 2) && (!($year & 3)))
	{
		# Extra day for Feb. 29
		$result += (24*60*60);
	}

	# Adjust for date of current month
	$result += (24*60*60 * ($date-1));

	return $result;
}
